/*
 * Copyright (C) 2011 Tal Aloni <tal.aloni.il@gmail.com>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ntddk.h>
#include <initguid.h>
#include <wdmsec.h>
#include <ntdddisk.h>
#include "nicbootconf.h"
#include "registry.h"

static NTSTATUS store_network_adapter_parameters (HANDLE deviceInstanceKey, LPWSTR classInstanceID) 
{
	LPCWSTR NetAdapterClassGUID = L"{4D36E972-E325-11CE-BFC1-08002BE10318}";
	
	NTSTATUS status;
	LPWSTR deviceDesc;
	LPWSTR service;
	LPWSTR newDriver;

	/* read existing keys */
	status = reg_fetch_sz ( deviceInstanceKey, L"Service", &service );
	if (status != STATUS_OBJECT_NAME_NOT_FOUND)
	{
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Failed to read Service entry value: %x\n", status );
		}
		else
		{
			DbgPrint ( "[VERBOSE] NIC service: %S\n",  service );
			ExFreePool ( service );
		}
	}
	/* finished reading existing keys */
	
	status = reg_fetch_sz ( deviceInstanceKey, L"DeviceDesc", &deviceDesc );
	if (status != STATUS_OBJECT_NAME_NOT_FOUND)
	{
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Failed to read DeviceDesc entry value: %x\n", status );
		}
		else
		{
			DbgPrint ( "[VERBOSE] NIC description: %S\n",  deviceDesc );
			ExFreePool ( deviceDesc );
		}
	}
	else
	{
		// "DeviceDesc" will usually be present, but otherwise we have to set it, because it's presence is critical for some reason.
		status = reg_store_sz ( deviceInstanceKey, L"DeviceDesc", L"Ethernet Controller" );
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Could not write \"DeviceDesc\". NTSTATUS: %x\n", status );
			return status;
		}
	}
	
	newDriver = (LPWSTR)ExAllocatePoolWithTag ( NonPagedPool, (wcslen(NetAdapterClassGUID) + 6) * 2, NICBOOTCONF_POOL_TAG );
	if ( ! newDriver ) 
	{
		DbgPrint ( "ExAllocatePoolWithTag failed to allocate memory for newDriver.\n");
		return STATUS_MEMORY_NOT_ALLOCATED;
	}
	wcsncpy(newDriver, NetAdapterClassGUID, wcslen(NetAdapterClassGUID) + 1); 	// copy the null termination too
	wcsncat(newDriver, L"\\", 2); 												// copy the null termination too
	wcsncat(newDriver, classInstanceID, wcslen(classInstanceID) + 1); 			// copy the null termination too
	
	status = reg_store_sz ( deviceInstanceKey, L"Driver", newDriver);
	if ( ! NT_SUCCESS ( status ) )
	{
		ExFreePool ( newDriver );
		DbgPrint ( "Could not update registry value \"Driver\". NTSTATUS: %x\n", status );
		return status;
	}
	
	DbgPrint ("Registry value \"Driver\" has been set to \"%S\"\n",  newDriver);
	ExFreePool ( newDriver );
	
	return status;
}

static NTSTATUS find_class_instance_id_for_network_adapter ( LPWSTR deviceID, ULONG matchIndex, LPWSTR * classInstanceId) 
{
	LPCWSTR NetAdapterClassGUID = L"{4D36E972-E325-11CE-BFC1-08002BE10318}"; // NetworkAdapterClassGUID
	LPCWSTR key_name_prefix = ( L"\\Registry\\Machine\\SYSTEM\\CurrentControlSet\\"
								L"Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318}"); 
								
	HANDLE reg_key;
	NTSTATUS status;
	ULONG subKeyCount;
	ULONG i;
	LPWSTR subKeyName;
	HANDLE sub_key;
	
	LPWSTR matchingDeviceId;
	LPWSTR currentDeviceID;
	ULONG matchCount = 0;

	status = reg_open ( &reg_key, key_name_prefix, NULL );
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Could not open registry key \"%S\". NTSTATUS: %x\n", key_name_prefix, status );
		return status;
	}
	
	status = reg_fetch_subkey_count(reg_key, &subKeyCount);
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Failed to fetch subkey count for registry key \"%S\". NTSTATUS: %x\n", key_name_prefix, status );
		reg_close(reg_key);
		return status;
	}
	
	DbgPrint ( "Looking for matching class instance to device-id \"%S\"\n", deviceID );	
	for(i=0; i < subKeyCount; i++)
	{
		status = reg_fetch_subkey_name(reg_key, i, &subKeyName);
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Failed to fetch the name of subkey number %d for registry key \"%S\". NTSTATUS: %x\n", i, key_name_prefix, status );
			reg_close(reg_key);
			return status;
		}
		
		status = reg_open ( &sub_key, key_name_prefix, subKeyName, NULL );
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Could not open registry key \"%S\\%S\". NTSTATUS: %x\n", key_name_prefix, subKeyName, status );
			reg_close(reg_key);
			return status;
		}
		
		/* read class instance registry values */
		status = reg_fetch_sz ( sub_key, L"MatchingDeviceId", &matchingDeviceId );	
		if (status == STATUS_OBJECT_NAME_NOT_FOUND) // no MatchingDeviceId present
		{
			DbgPrint ( "[VERBOSE] Skipping class instance \"%S\\%S\" without registry value \"MatchingDeviceId\".\n", NetAdapterClassGUID, subKeyName);
			reg_close(sub_key);
			continue;
		}
		else if ( !NT_SUCCESS ( status ) )
		{
			DbgPrint ( "[VERBOSE] Unexpected error. Could not read registry value \"MatchingDeviceId\" from registry key \"%S\\%S\". NTSTATUS: %x\n", key_name_prefix, subKeyName, status );
			reg_close(sub_key);
			continue;
		}
		else
		{
			DbgPrint ( "Class instance \"%S\\%S\": \"MatchingDeviceId\" value is \"%S\"\n",  NetAdapterClassGUID, subKeyName, matchingDeviceId );
			// matchingDeviceId is actually an hardwareID! (has the form of enumerator\deviceID)
			currentDeviceID = matchingDeviceId;
			
			// remove the enumerator name from currentDeviceID
			// _wcsnicmp is case-insensitive
			while (_wcsnicmp(currentDeviceID, L"\\", 1) != 0 && _wcsnicmp(currentDeviceID, L"\0", 1) != 0)
			{
				currentDeviceID++;
			}
			
			if (_wcsnicmp(currentDeviceID, L"\\", 1) == 0)
			{
				currentDeviceID++;
			}
			DbgPrint ( "Class instance \"%S\\%S\": deviceID is \"%S\"\n",  NetAdapterClassGUID, subKeyName, currentDeviceID );
			
			// Note: &REV can be changed as a result of firmware upgrade
			if (_wcsnicmp(deviceID, currentDeviceID, 17) == 0) // We only care about VEN_XXXX&DEV_XXXX (we ignore &SUBSYS and &REV)
			{
				// deviceID starts with currentDeviceID:
				// Note: typical deviceID    : VEN_8086&DEV_100F&SUBSYS_075015AD&REV_01
				//       typical currentDeviceID : ven_8086&dev_100f or ven_10ec&dev_8168&rev_06
				DbgPrint ( "Class instance \"%S\\%S\" is matching DeviceID \"%S\"\n",  NetAdapterClassGUID, subKeyName, deviceID );
				if (matchCount == matchIndex)
				{
					*classInstanceId = subKeyName;
					ExFreePool(matchingDeviceId);
					reg_close ( sub_key );
					reg_close ( reg_key );
					return STATUS_SUCCESS;
				}
				else
				{
					matchCount++;
				}
			}
			else
			{
				DbgPrint ( "No match was found for DeviceID \"%S\"\n", deviceID );
			}
			ExFreePool(matchingDeviceId);
		}
		reg_close ( sub_key );
		ExFreePool ( subKeyName );
	}
	
	
	reg_close ( reg_key );
	
	return STATUS_NOT_FOUND;
}

static NTSTATUS enumerate_devices(LPWSTR enumerator, LPWSTR deviceID)
{
	LPCWSTR NetAdapterClassGUID = L"{4D36E972-E325-11CE-BFC1-08002BE10318}";
	LPCWSTR key_name_prefix = L"\\Registry\\Machine\\SYSTEM\\CurrentControlSet\\Enum";
	
	NTSTATUS status;
	HANDLE device_key;
	ULONG subKeyCount;
	ULONG i;
	LPWSTR deviceInstanceId;
	HANDLE device_instance_key;
	LPWSTR classGUID;
	LPWSTR driver;
	LPWSTR classInstanceId;
	ULONG matchIndex = 0;
	
	status = reg_open ( &device_key, key_name_prefix, enumerator, deviceID, NULL );
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Could not open device instance key \"%S\\%S\". NTSTATUS: %x\n", enumerator, deviceID, status );
		return status;
	}
	
	status = reg_fetch_subkey_count(device_key, &subKeyCount);
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Failed to fetch subkey count for registry key \"%S\\%S\\%S\". NTSTATUS: %x\n", key_name_prefix, enumerator, deviceID, status );
		reg_close(device_key);
		return status;
	}
		
	for(i=0; i < subKeyCount; i++)
	{
		status = reg_fetch_subkey_name(device_key, i, &deviceInstanceId);
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Failed to fetch the name of subkey number %d for registry key \"%S\\%S\\%S\". NTSTATUS: %x\n", i, key_name_prefix, enumerator, deviceID, status );
			reg_close(device_key);
			return status;
		}
		
		status = reg_open ( &device_instance_key, key_name_prefix, enumerator, deviceID, deviceInstanceId, NULL );
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Could not open registry key \"%S\\%S\\%S\\%S\". %x\n", key_name_prefix, enumerator, deviceID, deviceInstanceId, status );
			reg_close(device_key);
			ExFreePool ( deviceInstanceId );
			return status;
		}
		
		/* read existing values */
		status = reg_fetch_sz ( device_instance_key, L"ClassGUID", &classGUID );	
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "[VERBOSE] device instance \"%S\\%S\" has no ClassGUID registry entry: %x\n", deviceID, deviceInstanceId, status );
			reg_close(device_instance_key);
			continue;
		}
		
		DbgPrint ( "[VERBOSE] device instance \"%S\\%S\" has the following ClassGUID: %S\n", deviceID, deviceInstanceId, classGUID );
		// _wcsnicmp is case-insensitive
		if (_wcsnicmp(classGUID, NetAdapterClassGUID, wcslen(NetAdapterClassGUID)) == 0)
		{
			// device has the ClassGUID of network adapter
			DbgPrint ( "device instance \"%S\\%S\\%S\" is a network adapter (matching Device Class GUID)\n", enumerator, deviceID, deviceInstanceId);
			status = reg_fetch_sz ( device_instance_key, L"Driver", &driver );	
			if (status == STATUS_OBJECT_NAME_NOT_FOUND) // NIC with no driver entry, this is what we're looking for
			{
				DbgPrint ( "Network adapter is not configured. searching for network adapter class instance matching DeviceID \"%S\"\n", deviceID);
				status = find_class_instance_id_for_network_adapter(deviceID, matchIndex, &classInstanceId);
				if (NT_SUCCESS(status)) // found matching network adapter class instance
				{
					DbgPrint ( "Asssigning class instance \"%S\\%S\" to device instance \"%S\\%S\"\n", NetAdapterClassGUID, classInstanceId, deviceID, deviceInstanceId);
					status = store_network_adapter_parameters(device_instance_key, classInstanceId);
					ExFreePool(classInstanceId);
					if ( ! NT_SUCCESS ( status ) )
					{
						DbgPrint ( "store_network_adapter_parameters failed to update device registry key \"%S\\%S\". NTSTATUS: %x\n", deviceID, deviceInstanceId, status);
						reg_close(device_key);
						ExFreePool ( deviceInstanceId );
						reg_close(device_instance_key);
						ExFreePool(classGUID);
						return status;
					}
					matchIndex++;
				}
				else if (status == STATUS_NOT_FOUND)
				{
					DbgPrint ( "[VERBOSE] Could not find class instance ID matching device instance \"%S\\%S\".\n", deviceID, deviceInstanceId );
				}
				else
				{
					DbgPrint ( "[VERBOSE] find_class_instance_id_for_network_adapter failed with unexpected error. NTSTATUS: %x\n", status );
				}
			}
			else if (NT_SUCCESS ( status ))
			{
				DbgPrint ( "device instance \"%S\\%S\\%S\" is already configured.\n", enumerator, deviceID, deviceInstanceId);
			}
			else
			{
				DbgPrint ( "[VERBOSE] Unexpected error. failed to fetch registry string value \"Driver\" from key \"%S\\%S\\%S\\%S\". NTSTATUS: %x", key_name_prefix, enumerator, deviceID, deviceInstanceId, status);
				reg_close(device_key);
				ExFreePool ( deviceInstanceId );
				reg_close(device_instance_key);
				ExFreePool(classGUID);
				return status;
			}
		}
		else
		{
			DbgPrint ( "[VERBOSE] device instance \"%S\\%S\\%S\" is NOT a network adapter\n", enumerator, deviceID, deviceInstanceId);
		}
		ExFreePool ( deviceInstanceId );
		reg_close(device_instance_key);
		ExFreePool(classGUID);
	}
	
	reg_close(device_key);
	
	return STATUS_SUCCESS;
}

static NTSTATUS enumerate_bus(LPWSTR enumerator)
{
	LPCWSTR key_name_prefix = L"\\Registry\\Machine\\SYSTEM\\CurrentControlSet\\Enum";

	NTSTATUS status;
	HANDLE reg_key;
	ULONG subKeyCount;
	ULONG i;
	LPWSTR subKeyName;
	
	status = reg_open ( &reg_key, key_name_prefix, enumerator, NULL );
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Failed to open registry key \"%S\\%S\". NTSTATUS: %x\n", key_name_prefix, enumerator, status );
		return status;
	}
	
	status = reg_fetch_subkey_count(reg_key, &subKeyCount);
	if ( ! NT_SUCCESS ( status ) )
	{
		DbgPrint ( "Failed to fetch subkey count for registry key \"%S\\%S\". NTSTATUS: %x\n", key_name_prefix, enumerator, status );
		reg_close(reg_key);
		return status;
	}
	
	for(i=0; i < subKeyCount; i++)
	{
		status = reg_fetch_subkey_name(reg_key, i, &subKeyName);
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "Failed to fetch the name of subkey number %d for registry key \"%S\\%S\". NTSTATUS: %x\n", i, key_name_prefix, enumerator, status );
			reg_close(reg_key);
			return status;
		}
		DbgPrint("[VERBOSE] Enumerated subkey: %S\n", subKeyName);
		
		status = enumerate_devices(enumerator, subKeyName);
		if ( ! NT_SUCCESS ( status ) )
		{
			DbgPrint ( "enumerate_devices failed for subkey \"%S\\%S\". NTSTATUS: %x\n", enumerator, subKeyName, status );
			reg_close(reg_key);
			ExFreePool ( subKeyName );
			return status;
		}
		ExFreePool ( subKeyName );
	}
	reg_close(reg_key);
	
	return STATUS_SUCCESS;
}

/**
 * Driver entry point
 *
 * @v DriverObject	Driver object
 * @v RegistryPath	Driver-specific registry path
 * @ret ntstatus	NT status
 */
NTSTATUS DriverEntry ( IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS status;
	LPWSTR enumerator = L"PCI";

	DbgPrint ( "NIC Boot Configuration Driver initializing\n" );
	
	status = enumerate_bus(enumerator);
	if ( ! NT_SUCCESS ( status ) ) 
	{
		DbgPrint ( "enumerate_bus returned NTSTATUS: %x\n", status );
	}
	
	DbgPrint ( "NIC Boot Configuration Driver finalizing\n" );
	UNREFERENCED_PARAMETER(DriverObject);
	UNREFERENCED_PARAMETER(RegistryPath);
	return status;
}
